;;; $Id: image.scm 108 2005-12-16 23:55:04Z rlpm $

;;; An Image Processing Library for MzScheme
;;; Author: Rory McGuire <rlpm at cs dot unm dot edu>

;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 2 of the
;;; License, or (at your option) any later version.
     
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.

;;; You can obtain a copy of the GPL at:
;;;   http://www.fsf.org/copyleft/gpl.html.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; This file requires functions defined in helpers.scm and matrix.scm.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; The interface for the image processing library for MzScheme was
;;; directly derived from, and in many cases identical to, that of the
;;; primitive image data type and image processing extensions in UNM
;;; Scheme.  UNM Scheme is an implementation of the Scheme programming
;;; language written by Professor Lance Williams of the University
;;; of New Mexico.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; <struct:image>
;   type : 'grey | 'complex | 'color
;   rows : int
;   cols : int
;   array : #(#(<row0>)..#(<rowN>))
(define-values (struct:image
                image-internal-make
                image?
                image-internal-ref
                image-internal-set!)
  (make-struct-type 'image #f 4 0 #f () (make-inspector)))

(define image-internal-type
  (make-struct-field-accessor image-internal-ref 0 'type))

(define (grey-image? img)
  (and (image? img) (eq? (image-internal-type img) 'grey)))

(define gray-image? grey-image?)

(define (complex-image? img)
  (and (image? img) (eq? (image-internal-type img) 'complex)))

(define (color-image? img)
  (and (image? img) (eq? (image-internal-type img) 'color)))

(define image-rows
  (make-struct-field-accessor image-internal-ref 1 'rows))

(define image-cols
  (make-struct-field-accessor image-internal-ref 2 'cols))

(define image->vector
  (make-struct-field-accessor image-internal-ref 3 'vector))

(define (internal-color-valok x)
  (and (pair? x) (= 3 (length x)) ((list-all real?) x)))

(define (internal-color-valcheck s x)
  (if (internal-color-valok x)
      x
      (error (string-append s ": expects argument of type <rgb val list>:"
                            "given") x)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; n = rows
; m = cols
(define (make-image n m . f)
  (define (check-int s x)
    (if (and (integer? x) (> x 0))
        x
        (error (string-append "make-image: expects type"
                              " <non-negative exact integer>"
                              " as "
                              s
                              " argument, given:")
               x)))
  (let* ((rows (check-int "1st" n))
         (cols (check-int "2nd" m))
         (f (cond ((null? f) (lambda (i j) 0))
                  ((and (null? (cdr f)) (procedure? (car f))) (car f))
                  (else (error (string-append "make-image: expects type "
                                              "<procedure> as optional 3rd "
                                              "argument, given:")
                               (car f) (cdr f))))))
    (define (loop i j type tmp acc)
      (cond ((= i rows) (list type (reverse acc)))
            ((= j cols) (loop (+ i 1) 0 type () (cons (reverse tmp) acc)))
            (else
             (let ((val (f i j)))
               (cond ((eq? 'color type)
                      (loop i (+ 1 j) 'color
                            (cons (internal-color-valcheck "make-image" val)
                                  tmp) acc))
                     ((not (number? val))
                      (error (string-append
                              "make-image: value at "
                              (number->string i)
                              ","
                              (number->string j)
                              " is not a number")))
                     ((eq? 'complex type)
                      (loop i (+ 1 j) type (cons val tmp) acc))
                     ((not (real? val))
                      (loop i (+ 1 j) 'complex (cons val tmp) acc))
                     (else
                      (loop i (+ 1 j) (if (real? val) 'grey 'complex)
                            (cons val tmp) acc)))))))
    (define (finalize-image type ls)
      (image-internal-make type rows cols
                           (list->vector (map list->vector ls))))
    (define (initial-type x)
      (if (number? x)
          (if (real? x) 'grey 'complex)
          (if (internal-color-valok x)
              'color
              (error "make-image: value at 0,0 not a number or rgb list"))))
    (let ((initial (f 0 0)))
      (apply finalize-image
             (loop 0 1 (initial-type initial) (list initial) ())))))

; helper to complain if an expected object is not an image
(define (image-typecheck s i)
  (if (not (image? i))
      (error (string-append s ": expects argument of type <image>: given") i)
      i))

(define (image-ref img i j)
  (let ((img (image-typecheck "image-ref" img)))
    (cond ((or (not (integer? i)) (< i 0) (>= i (image-rows img)))
           (error "image-ref: row index out of range" i))
          ((or (not (integer? j)) (< j 0) (>= j (image-cols img)))
           (error "image-ref: col index out of range" j))
          (else
           (vector-ref (vector-ref (image->vector img) i) j)))))

(define (image-map f img)
  (let ((img (image-typecheck "image-map" img)))
    (make-image (image-rows img) (image-cols img)
                (lambda (i j) (f (image-ref img i j))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (vector->image v)
  (define (check x)
    (or (number? x) (internal-color-valcheck "vector->image" x)))
  (if (not (and (vector? v)
                (let ((inner (vector-ref v 0)))
                  (and (vector? inner)
                       (let ((cols (vector-length inner)))
                         ((list-all
                           (lambda (x)
                             (and (vector? x)
                                  (= cols (vector-length x))
                                  ((list-all check) (vector->list x)))))
                          (vector->list v)))))))
      (error (string-append "vector->image: expects vector of "
                            "same-sized number vector"))
      (make-image (vector-length v)
                  (vector-length (vector-ref v 0))
                  (lambda (i j) (vector-ref (vector-ref v i) j)))))

(define image->array image->vector)

(define array->image vector->image)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (image-set! img i j x)
  (image-typecheck "image-set!" img)
  (cond ((or (not (integer? i)) (< i 0) (>= i (image-rows img)))
         (error "image-set!: row index out of range" i))
        ((or (not (integer? j)) (< j 0) (>= j (image-cols img)))
         (error "image-set!: col index out of range" j))
        ((color-image? img)
         (vector-set! (vector-ref (image->vector img) i) j
                      (internal-color-valcheck "image-set! (color)" x)))
        ((not (number? x))
         (error "image-set!: value not a number: " x))
        (else
         (if (not (real? x)) (image-internal-set! img 0 'complex))
         (vector-set! (vector-ref (image->vector img) i) j x))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (image-crop img i j n m)
  (let ((img (image-typecheck "image-crop" img)))
    (make-image n m (lambda (i_ j_) (image-ref img (+ i i_) (+ j j_))))))

(define (image-pad img n m)
  (let ((black (if (color-image? (image-typecheck "image-crop" img))
                   '(0 0 0)
                   0)))
    (make-image n m
                (lambda (i j)
                  (if (or (>= i (image-rows img)) (>= j (image-cols img)))
                      black
                      (image-ref img i j))))))

(define (image-normalize img)
  (if (or (not (image? img)) (complex-image? img))
      (error "image-normalize: expects grey or color image"))
  (let* ((initial (image-ref img 0 0))
         (minval (if (pair? initial) (apply min initial) initial))
         (maxval (if (pair? initial) (apply max initial) initial))
         (unused 
          (do ((x 0 (+ x 1))) ((= x (image-rows img)) (void))
            (do ((y 0 (+ y 1))) ((= y (image-cols img)) (void))
              (let* ((cur (image-ref img x y))
                     (cur (if (pair? cur) cur (list cur))))
              (set! minval (apply min minval cur))
              (set! maxval (apply max maxval cur))))))
         (max_min (- maxval minval))
         (f (lambda (x)
              (if (= 0 max_min)
                  0
                  (exact->inexact (/ (- x minval) max_min)))))
         (fl (lambda (x) (map f x))))
    (image-map (if (color-image? img) fl f) img)))

(define (image-normalize256 img)
  (let* ((f (lambda (x) (inexact->exact (round (* 255 x)))))
         (fl (lambda (x) (map f x))))
  (image-map (if (color-image? img) fl f) (image-normalize img))))

(define (image-shrink img t)
  (define (c x)
    (let ((mag (magnitude x))
          (ang (angle x)))
      (* (if (< mag t) 0 (- mag t)) (exp (* 0-i ang)))))
  (define (g x)
    (if (< (abs x) t) 0 ((if (< x 0) + -) x t)))
  (cond ((complex-image? img) (image-map c img))
        ((grey-image? img) (image-map g img))
        (else (error "image-normalize: expects grey or complex image"))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (image-transpose img)
  (make-image (image-cols img) (image-rows img)
              (lambda (i j) (image-ref img j i))))

(define (image->matrix img)
  (map vector->list (vector->list (image->vector img))))

(define (matrix->image M)
  (vector->image (list->vector (map list->vector M))))

(define (image-product i0 i1)
  (matrix->image (matrix-product (image->matrix i0) (image->matrix i1))))

(define (image-binop f)
  (lambda (img1 img2)
    (make-image (image-rows img1) (image-cols img1)
                (lambda (i j)
                  (f (image-ref img1 i j) (image-ref img2 i j))))))

(define image_+ (image-binop +))
(define image_- (image-binop -))
(define image_* (image-binop *))
(define image_/ (image-binop /))
(define image_log (lambda (img) (image-map log img)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (fold-right f img)
  (define (fold-right-helper a . rest)
    (if (null? rest)
        a
        (f a (apply fold-right-helper rest))))
  (matrix->image
   (map list 
        (map (lambda (x) (apply fold-right-helper x)) (image->matrix img)))))

(define (fold-down f img)
  (image-transpose (fold-right f (image-transpose img))))

(define (fold f img)
  (image-ref (fold-down f (fold-right f img)) 0 0))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (image-histogram256 img)
  (areas (image-normalize256 img)))

(define (areas img)
  (let* ((n (image-rows img))
         (m (image-cols img))
         (h (make-vector (+ 1 (fold max img)))))
    (do ((n_ 0 (+ 1 n_))) ((= n_ n) h)
      (do ((m_ 0 (+ 1 m_))) ((= m_ m)())
        (let ((i (image-ref img n_ m_)))
          (vector-set! h i (+ 1 (vector-ref h i))))))))

(define (_edge? img i j)
  (define (image-ref-torus i j)
    (image-ref img
               (modulo i (image-rows img))
               (modulo j (image-cols img))))
  (let ((cur (image-ref img i j)))
    (and (not (= 0 cur))
         (or (= 0 (image-ref-torus (- i 1) j))
             (= 0 (image-ref-torus (+ i 1) j))
             (= 0 (image-ref-torus i (- j 1)))
             (= 0 (image-ref-torus i (+ j 1)))))))

(define (perimeters img)
  (let ((vec (make-vector (+ 1 (fold max img)))))
    (do ((i 0 (+ 1 i))) ((= i (image-rows img)) vec)
      (do ((j 0 (+ 1 j))) ((= j (image-cols img))())
        (let ((cur (image-ref img i j)))
          (if (_edge? img i j)
              (let ((cur (image-ref img i j)))
                (vector-set! vec cur (+ 1 (vector-ref vec cur)))
                (vector-set! vec 0 (+ 1 (vector-ref vec 0))))))))))

(define (outline img . x)
  (let ((x (if (null? x) 1 (car x)))
        (y (if (or (null? x) (null? (cdr x))) 0 (cadr x))))
  (make-image (image-rows img) (image-cols img)
              (lambda (i j)
                (if (_edge? img i j) x y)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (downsample-rows img)
  (let ((img (image-typecheck "downsample-rows" img)))
    (make-image (quotient (image-rows img) 2) (image-cols img)
                (lambda (i j) (image-ref img (* i 2) j)))))

(define (downsample-cols img)
  (let ((img (image-typecheck "downsample-cols" img)))
    (make-image (image-rows img) (quotient (image-cols img) 2)
                (lambda (i j) (image-ref img i (* 2 j))))))

(define (downsample img)
  (let ((img (image-typecheck "downsample" img)))
        (downsample-rows (downsample-cols img))))

(define (upsample-rows img)
  (let* ((img (image-typecheck "upsample-rows" img))
         (black (if (color-image? img) '(0 0 0) 0)))
    (make-image (* (image-rows img) 2) (image-cols img)
                (lambda (i j)
                  (if (= 0 (modulo i 2))
                      (image-ref img (quotient i 2) j)
                      black)))))

(define (upsample-cols img)
  (let* ((img (image-typecheck "upsample-cols" img))
         (black (if (color-image? img) '(0 0 0) 0)))
    (make-image (image-rows img) (* (image-cols img) 2)
                (lambda (i j)
                  (if (= 0 (modulo j 2))
                      (image-ref img i (quotient j 2))
                      black)))))

(define (upsample img)
  (let ((img (image-typecheck "upsample" img)))
    (upsample-rows (upsample-cols img))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (internal-left-to-right img img2)
  (image-typecheck "left-to-right" img)
  (image-typecheck "left-to-right" img2)
  (if (or (and (color-image? img) (not (color-image? img2)))
          (and (not (color-image? img)) (color-image? img2))
          (not (= (image-rows img) (image-rows img2))))
      (error (string-append "left-to-right: expects same-type images with "
                            "equal number of rows"))
      (make-image (image-rows img)
                  (+ (image-cols img) (image-cols img2))
                  (lambda (i j)
                    (let* ((first (< j (image-cols img)))
                           (j (if first j (- j (image-cols img))))
                           (img (if first img img2)))
                      (image-ref img i j))))))

(define (left-to-right img . rest)
  (if (null? rest)
      img
      (apply left-to-right
             (internal-left-to-right img (car rest))
             (cdr rest))))

(define (top-to-bottom img . rest)
  (if (null? rest)
      img
      (apply top-to-bottom
             (image-transpose
              (internal-left-to-right (image-transpose img)
                                      (image-transpose (car rest))))
             (cdr rest))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-filter n m . f)
  (let ((f (cond ((null? f) (lambda (i j) 0))
                 ((and (null? (cdr f)) (procedure? (car f))) (car f))
                 (else (error (string-append "make-filter: expects type "
                                              "<procedure> as optional 3rd "
                                              "argument, given:")
                               (car f) (cdr f)))))
        (n_2 (/ n 2))
        (m_2 (/ m 2)))
    (make-image n m (lambda (i j) (f (if (< i n_2) i (- i n))
                                     (if (< j m_2) j (- j m)))))))
  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define 180_deg (* 2 (acos 0)))
(define 360_deg (* 2 180_deg))
(define (complex->color img)
  (let* ((ang (image-map (o (lambda (x) (if (= x 1) 0 x))
                            (lambda (x) (if (< x 0) (+ 1 x) x))
                            (lambda (x) (/ x 360_deg)))
                         (image-angle img)))
         (mag (image-magnitude img))
         (mxm (fold max mag))
         (mag (image-map (lambda (x) (/ x mxm)) mag)))
    (make-image (image-rows img) (image-cols img)
                (lambda (i j)
                  (_hsv->rgb (image-ref ang i j)
                             1
                             (image-ref mag i j))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; parse a PGMA file into an image object
(define (read-image name)
    ; read a PGMA file, using parse-string to ignore # comments
    ; returns a list of the next line of the file
    (define (read-pgma file)
      (if (eq? #\# (peek-char file))
          (begin (read-line file 'any) (read-pgma file))
          (read file)))
    (let* ((file (open-input-file name))
           (magic (read-line file 'any))
           (unused (if (not (or (equal? "P2" magic) (equal? "P3" magic)))
                       (begin (close-input-port file)
                              (error "Not a PGMA or PPMA file: " name))))
           (color (equal? magic "P3"))
           (cols (read-pgma file))
           (rows (read-pgma file))
           (max (read-pgma file))
           (img (make-image rows cols
                            (if color
                                (lambda (i j) (list (read-pgma file)
                                                    (read-pgma file)
                                                    (read-pgma file)))
                                (lambda (i j) (read-pgma file)))))
           (unused (close-input-port file)))
      img))

(define (write-image img name)
  (define (f n file)
    (if (< n 100) (display " " file))
    (if (< n 10) (display " " file))
    (display n file)
    (display " " file))
  (let* ((img (image-normalize256 (if (complex-image? img)
                                      (complex->color img)
                                      img)))
         (f (if (color-image? img)
                (lambda (x file) (map (lambda (x) (f x file)) x))
                f))
         (m (if (color-image? img) 5 16))
         (rows (image-rows img))
         (cols (image-cols img)))
    (define (loop file)
      (display (if (color-image? img) "P3" "P2") file)
      (newline file)
      (display "# CREATOR: IMAGE for MzScheme $Revision: 108 $" file)
      (newline file)
      (display cols file)
      (display " " file)
      (display rows file)
      (newline file)
      (display "255" file)
      (do ((i 0 (+ i 1))(n 0 (+ cols n))) ((= i rows) (void))
        (do ((j 0 (+ j 1))(n n (+ n 1))) ((= j cols) (void))
          (if (= 0 (modulo n m)) (newline file) (display " " file))
          (f (image-ref img i j) file))))
    (if (output-port? name) (loop name)
        (begin
          (if (file-exists? name) (delete-file name))
          (call-with-output-file name loop)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define display-image-exe "/usr/bin/display")
(define display-image-stdin #t)
(define display-image-dotitle #t)
(define display-image-tmpdir "/tmp/display-image/")
(define display-image-idx 0)
(define display-image-total 10)

(define (display-image img)
  (image-typecheck "display-image" img)
  (let* ((tag (string-append "#<image ("
                             (symbol->string (image-internal-type img))
                             "): rows = "
                             (number->string (image-rows img))
                             " cols = "
                             (number->string (image-cols img))
                             ">"))
         (unused (begin (display tag)(newline)))
         (filename
          (if display-image-stdin
              "-"
              (let ((f (string-append display-image-tmpdir
                                      "display-image-"
                                      (number->string display-image-idx))))
                (set! display-image-idx
                      (modulo (+ 1 display-image-idx) display-image-total))
                (write-image img f)
                f))))
    (define-values (pid p1 p2 p3)
      (apply subprocess (current-output-port) #f (current-error-port)
                  display-image-exe
                  (if display-image-dotitle
                      (append (list "-title" tag) (list filename))
                      (list filename))))
    (if display-image-stdin (write-image img p2))
    (close-output-port p2)))

(define (image-repl)
  (let ((orig-print (current-print)))
    (parameterize ((current-print
                    (lambda (obj)
                      (if (image? obj) (display-image obj) (orig-print obj)))))
                  (display (string-append "Starting IMAGE for MzScheme "
                                          "hotwired Read-Eval-Print Loop"))
                  (newline)
                  (read-eval-print-loop)))
    (newline))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (label-newset sets)
  (list->vector (append (vector->list sets) (list (vector-length sets)))))

(define (label-unify sets greater lesser)
  (if (< greater lesser)
      (label-unify sets lesser greater)
      (let ((vec (list->vector (vector->list sets)))
            (new (vector-ref sets lesser))
            (old (vector-ref sets greater)))
        (if (< new old) (vector-set! vec greater new))
        vec)))

(define (label-compress sets)
  (define (helper sets i n acc)
    (if (null? sets)
        (reverse acc)
        (let* ((cur (car sets))
               (newn (+ n (if (= cur i) 1 0))))
          (helper (cdr sets)
                  (+ 1 i)
                  newn
                  (cons (if (= n newn)
                            (list-ref acc (- (length acc) cur 1))
                            newn)
                        acc)))))
  (list->vector (helper (vector->list sets) 0 -1 ())))

(define (label img)
  (let* ((rows (image-rows img))
         (cols (image-cols img))
         (tmp (make-image rows cols))
         (sets #(0)))
    (define (image-ref-helper img i j)
      (if (or (< i 0) (< j 0))
          0
          (image-ref img i j)))
    (define (sethelper sets i j)
      (if (= 0 (image-ref img i j))
          (cons 0 sets)
          (let ((above (image-ref-helper tmp (- i 1) j))
                (left (image-ref-helper tmp i (- j 1))))
            (cond ((and (= 0 above) (= 0 left))
                   (cons (vector-length sets) (label-newset sets)))
                  ((= 0 above) (cons left sets))
                  ((or (= above left) (= 0 left)) (cons above sets))
                  (else (cons (if (< above left) above left)
                              (label-unify sets left above)))))))
    (define (helper i j sets)
      (cond ((= i rows) sets)
            ((= j cols) (helper (+ i 1) 0 sets))
            (else
             (let* ((vals (sethelper sets i j))
                    (sets (cdr vals))
                    (color (vector-ref sets (car vals))))
               (image-set! tmp i j color)
               (helper i (+ j 1) sets)))))
    (let* ((sets (helper 0 0 sets))
           (sets (label-compress sets)))
      (make-image rows cols
                  (lambda (i j) (vector-ref sets (image-ref tmp i j)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (image-real-part img)
  (if (or (not (image? img)) (color-image? img))
      (error "image-real-part: expects non-color image as parameter, given:"
             img))
  (image-map real-part img))

(define (image-imag-part img)
  (if (or (not (image? img)) (color-image? img))
      (error "image-imag-part: expects non-color image, given:"
             img))
  (image-map imag-part img))

(define (image-complex img iimg)
  (if (or (not (grey-image? img))
          (not (grey-image? iimg))
          (not (= (image-rows img) (image-rows iimg)))
          (not (= (image-cols img) (image-cols iimg))))
      (error "image-complex: expects two grey images of same size")
      (make-image (image-rows img) (image-cols img)
                  (lambda (i j) (+ (image-ref img i j)
                                   (* 0+i (image-ref iimg i j)))))))

(define rectangular->complex-image image-complex)

(define (complex-image->rectangular img)
  (if (or (not (image? img)) (color-image? img))
      (error "complex-image->rectangular: expects non-color image")
      (list (image-real-part img) (image-imag-part img))))

(define (image-magnitude img)
  (if (or (not (image? img)) (color-image? img))
      (error "image-magnitude: expects non-color image")
      (image-map magnitude img)))

(define (image-angle img)
  (if (or (not (image? img)) (color-image? img))
      (error "image-angle: expects non-color image")
      (image-map (lambda (x) (if (= 0 x) 0 (angle x))) img)))

(define (complex-image->polar img)
  (if (or (not (image? img)) (color-image? img))
      (error "complex-image->polar: expects non-color image")
      (list (image-magnitude img) (image-angle img))))

(define (polar->complex-image aimg timg)
  (if (or (not (grey-image? aimg))
          (not (grey-image? timg))
          (not (= (image-rows aimg) (image-rows timg)))
          (not (= (image-cols aimg) (image-cols timg))))
      (error "polar->complex-image: expects two grey images of same size")
      (make-image (image-rows aimg) (image-cols aimg)
                  (lambda (i j) (* (image-ref aimg i j)
                                   (exp (* 0+i (image-ref timg i j))))))))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (color-image-red img)
  (if (not (color-image? img))
      (error ("color-image-red: expects color image"))
      (make-image (image-rows img) (image-cols img)
                  (lambda (i j) (car (image-ref img i j))))))

(define (color-image-green img)
  (if (not (color-image? img))
      (error ("color-image-green: expects color image"))
      (make-image (image-rows img) (image-cols img)
                  (lambda (i j) (cadr (image-ref img i j))))))

(define (color-image-blue img)
  (if (not (color-image? img))
      (error ("color-image-blue: expects color image"))
      (make-image (image-rows img) (image-cols img)
                  (lambda (i j) (caddr (image-ref img i j))))))

(define (color-image->rgb img)  
  (if (not (color-image? img))
      (error "color-image->rgb: expects color image")
      (list (color-image-red img)
            (color-image-green img)
            (color-image-blue img))))

(define (color-image-intensity img)
  (if (not (color-image? img))
      (error "color-image->intensity: expects color image")
      (make-image (image-rows img) (image-cols img)
                  (lambda (i j)
                    (/ (apply + (image-ref img i j)) 3)))))

(define (rgb->color-image r g b)
  (if (or (not (grey-image? r)) (not (grey-image? g)) (not (grey-image? b))
          (not (= (image-rows r) (image-rows g) (image-rows b)))
          (not (= (image-cols r) (image-cols g) (image-cols b))))
      (error ("rgb->color-image: expects grey images of same size"))
      (make-image (image-rows r) (image-cols r)
                  (lambda (i j) (list (image-ref r i j)
                                      (image-ref g i j)
                                      (image-ref b i j))))))

(define (color->grey img)
  (if (not (color-image? img))
      (error ("color->grey: expects color image"))
      (image-normalize
       (make-image (image-rows img) (image-cols img)
                   (lambda (i j) (apply + (image-ref img i j)))))))
(define color->gray color->grey)

(define (grey->color img)
  (if (not (grey-image? img))
      (error ("grey->color: expects grey image"))
      (let ((img (image-normalize img)))
        (make-image (image-rows img) (image-cols img)
                    (lambda (i j)
                      (let ((v (image-ref img i j)))
                        (list v v v)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (_rgb->hsv r g b)
  (let* ((var_min (min r g b))
         (var_max (max r g b))
         (del_max (- var_max var_min)))
    (if (= del_max 0)
        (list 0 0 var_max)
        (let* ((s (/ del_max var_max))
               (del (lambda (x) (/ (+ (/ (- var_max x)
                                         6)
                                      (/ del_max 2))
                                   del_max)))
               (del_r (del r))
               (del_g (del g))
               (del_b (del b))
               (h (cond ((= r var_max) (- del_b del_g))
                        ((= g var_max) (+ 1/3 del_r (- del_b)))
                        (else (+ 2/3 del_g (- del_r)))))
               (h (if (< h 0) (+ h 1) (if (> h 1) (- h 1) h))))
          (list h s var_max)))))

(define (_hsv->rgb h s v)
  (if (= s 0)
      (list v v v)
      (let* ((var_h (* h 6))
             (var_i (floor var_h))
             (var_1 (* v (- 1 s)))
             (var_2 (* v (- 1 (* s (- var_h var_i)))))
             (var_3 (* v (- 1 (* s (- 1 (- var_h var_i)))))))
        (cond ((= var_i 0.0) (list v var_3 var_1))
              ((= var_i 1.0) (list var_2 v var_1))
              ((= var_i 2.0) (list var_1 v var_3))
              ((= var_i 3.0) (list var_1 var_2 v))
              ((= var_i 4.0) (list var_3 var_1 v))
              (else (list v var_1 var_2))))))

(define (rgb->hsv r g b)
  (if (or (not (grey-image? r)) (not (grey-image? g)) (not (grey-image? b))
          (not (= (image-rows r) (image-rows g) (image-rows b)))
          (not (= (image-cols r) (image-cols g) (image-cols b))))
      (error ("rgb->hsv: expects grey images of same size"))
      (color-image->rgb (make-image
                         (image-rows r)
                         (image-cols r)
                         (lambda (n m)
                           (let* ((r (image-ref r n m))
                                  (g (image-ref g n m))
                                  (b (image-ref b n m)))
                             (_rgb->hsv r g b)))))))

(define (hsv->color-image h s v)
  (if (or (not (grey-image? h)) (not (grey-image? s)) (not (grey-image? v))
          (not (= (image-rows h) (image-rows s) (image-rows v)))
          (not (= (image-cols h) (image-cols s) (image-cols v))))
      (error ("hsv->rgb: expects grey images of same size"))
      (make-image (image-rows h) (image-cols h)
                  (lambda (n m)
                    (_hsv->rgb (image-ref h n m)
                               (image-ref s n m)
                               (image-ref v n m))))))

(define (color-image->hsv img)
  (color-image->rgb
   (make-image (image-rows img)
               (image-cols img)
               (lambda (i j) (apply _rgb->hsv (image-ref img i j))))))

(define (hsv->rgb h s v)
  (color-image->rgb (hsv->color-image h s v)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (_rgb->hsl r g b)
  (let* ((var_min (min r g b))
         (var_max (max r g b))
         (del_max (- var_max var_min))
         (l (/ (+ var_max var_min) 2)))
    (if (= 0 del_max)
        (list 0 0 l)
        (let* ((s (/ del_max (if (< l 0.5)
                                 (+ var_max var_min)
                                 (- 2 var_max var_min))))
               (del (lambda (x) (/ (+ (/ (- var_max x) 6)
                                      (/ del_max 2))
                                   del_max)))
               (del_r (del r))
               (del_g (del g))
               (del_b (del b))
               (h (cond ((= var_max r) (- del_b del_g))
                        ((= var_max g) (+ 1/3 del_r (- del_b)))
                        (else (+ 2/3 del_g (- del_r)))))
               (h (if (< h 0) (+ h 1) (if (> h 1) (- h 1) h))))
          (list h s l)))))

(define (_hsl->rgb h s l)
  (if (= s 0)
      (list l l l)
      (let* ((v2 (if (< l 0.5)
                        (* l (+ 1 s))
                        (- (+ l s) (* s l))))
             (v1 (- (* 2 l) v2))
             (hue2rgb (lambda (vh)
                        (let* ((vh (if (< vh 0)
                                       (+ vh 1)
                                       (if (> vh 1) (- vh 1) vh))))
                          (cond ((< (* 6 vh) 1)
                                 (+ v1 (* (- v2 v1) 6 vh)))
                                ((< (* 2 vh) 1)
                                 v2)
                                ((< (* 3 vh) 2)
                                 (+ v1 (* (- v2 v1) (- 2/3 vh) 6)))
                                (else
                                 v1))))))
        (list
         (hue2rgb (+ h 1/3))
         (hue2rgb h)
         (hue2rgb (- h 1/3))))))

(define (rgb->hsl r g b)
  (if (or (not (grey-image? r)) (not (grey-image? g)) (not (grey-image? b))
          (not (= (image-rows r) (image-rows g) (image-rows b)))
          (not (= (image-cols r) (image-cols g) (image-cols b))))
      (error ("rgb->hsl: expects grey images of same size"))
      (color-image->rgb (make-image
                         (image-rows r)
                         (image-cols r)
                         (lambda (n m)
                           (let* ((r (image-ref r n m))
                                  (g (image-ref g n m))
                                  (b (image-ref b n m)))
                             (_rgb->hsl r g b)))))))

(define (hsl->color-image h s l)
  (if (or (not (grey-image? h)) (not (grey-image? s)) (not (grey-image? l))
          (not (= (image-rows h) (image-rows s) (image-rows l)))
          (not (= (image-cols h) (image-cols s) (image-cols l))))
      (error ("hsl->rgb: expects grey images of same size"))
      (make-image (image-rows h) (image-cols h)
                  (lambda (n m)
                    (_hsl->rgb (image-ref h n m)
                               (image-ref s n m)
                               (image-ref l n m))))))

(define (color-image->hsl img)
  (color-image->rgb
   (make-image (image-rows img)
               (image-cols img)
               (lambda (i j) (apply _rgb->hsl (image-ref img i j))))))

(define (hsl->rgb h s l)
  (color-image->rgb (hsl->color-image h s l)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define _360_deg (* 4 (acos 0)))
(define (_rgb->hsi r g b)
  (let* ((i (* 1/3 (+ r g b))))
    (if (= i 0) ; black
        (list 0 0 0)
        (let* ((s (- 1 (* 3 (min r g b) (/ (+ r g b)))))
               (r_g (- r g))
               (r_b (- r b))
               (g_b (- g b))
               (den (* 2 (sqrt (+ (* r_g r_g) (* r_b g_b))))))
          (if (= den 0) ; grey
              (list 0 0 i)
              (let* ((h (acos (/ (+ r_g r_b) den)))
                     (h (if (> b g) (- _360_deg h) h)))
                (list (/ h _360_deg) s i)))))))

(define _60_deg (/ _360_deg 6))
(define _120_deg (* _60_deg 2))
(define _240_deg (* _120_deg 2))
(define (_hsi->rgb h s i)
  (if (= s 0)
      (list i i i)
      (let* ((h (* _360_deg h))
             (v1 (/ (- 1 s) 3)))
        (define (val2 h s)
          (* 1/3 (+ 1 (/ (* s (cos h)) (cos (- _60_deg h))))))
        (define (norm x) (* 3 i x))
        (cond ((< h _120_deg)
               (let ((r (val2 h s)))
                 (map norm (list r (- 1 v1 r) v1))))
              ((< h _240_deg)
               (let ((g (val2 (- h _120_deg) s)))
                 (map norm (list v1 g (- 1 g v1)))))
              (else
               (let ((b (val2 (- h _240_deg) s)))
                 (map norm (list (- 1 v1 b) v1 b))))))))

(define (rgb->hsi r g b)
  (if (or (not (grey-image? r)) (not (grey-image? g)) (not (grey-image? b))
          (not (= (image-rows r) (image-rows g) (image-rows b)))
          (not (= (image-cols r) (image-cols g) (image-cols b))))
      (error ("rgb->hsi: expects grey images of same size"))
      (color-image->rgb (make-image
                         (image-rows r)
                         (image-cols r)
                         (lambda (n m)
                           (let* ((r (image-ref r n m))
                                  (g (image-ref g n m))
                                  (b (image-ref b n m)))
                             (_rgb->hsi r g b)))))))

(define (hsi->color-image h s i)
  (if (or (not (grey-image? h)) (not (grey-image? s)) (not (grey-image? i))
          (not (= (image-rows h) (image-rows s) (image-rows i)))
          (not (= (image-cols h) (image-cols s) (image-cols i))))
      (error ("hsi->rgb: expects grey images of same size"))
      (make-image (image-rows h) (image-cols h)
                  (lambda (n m)
                    (_hsi->rgb (image-ref h n m)
                               (image-ref s n m)
                               (image-ref i n m))))))

(define (color-image->hsi img)
  (let ((img (image-normalize img)))
    (color-image->rgb
     (make-image (image-rows img)
                 (image-cols img)
                 (lambda (i j) (apply _rgb->hsi (image-ref img i j)))))))

(define (hsi->rgb h s i)
  (color-image->rgb (hsi->color-image h s i)))

(define (huemap img)
  (if (color-image? img)
      (huemap (color->grey img))
      (let ((one (make-image (image-rows img) (image-cols img)
                             (lambda (i j) 1))))
      (hsi->color-image (image-normalize img) one one))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-hot-image img)
  (if (not (grey-image? img))
      (error "make-hot-image: expects grey image")
      (let* ((img (image-normalize img)))
        (make-image (image-rows img) (image-cols img)
                    (lambda (i j)
                      (let ((val (image-ref img i j)))
                        (cond ((< val 1/3) (list (* 3 val) 0 0))
                              ((< val 2/3) (list 1 (* 3 (- val 1/3)) 0))
                              (else (list 1 1 (* 3 (- val 2/3)))))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (image-rotate img)
  (make-image (image-cols img) (image-rows img)
              (lambda (i j) (image-ref img (- (image-rows img) j 1) i))))

(define (image-flip img)
  (make-image (image-rows img) (image-cols img)
              (lambda (i j) (image-ref img (- (image-rows img) i 1) j))))

(define (image-flop img)
  (make-image (image-rows img) (image-cols img)
              (lambda (i j) (image-ref img i (- (image-cols img) j 1)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (_discrete-convolution f g)
  (let* ((flen (vector-length f))
         (glen (vector-length g))
         (gz (quotient glen 2))
         (iter (reverse (iota0 glen))))
    (define (fref t)
      (vector-ref f (modulo t flen)))
    (list->vector
     (map (lambda (t)
            (apply + (map (lambda (x)
                            (* (fref (+ t (- gz x)))
                               (vector-ref g x))) iter)))
          (iota0 flen)))))

(define (convolve-rows img vec)
  (let ((iv (image->vector img)))
    (vector->image
     (list->vector
      (map (lambda (r)
             (_discrete-convolution (vector-ref iv r) vec))
           (iota0 (image-rows img)))))))

(define (convolve-cols img vec)
  (image-transpose (convolve-rows (image-transpose img) vec)))

(define (convolve img vec)
  (let* ((rows (image-rows img))
         (cols (image-cols img))
         (krows (vector-length vec))
         (kcols (vector-length (vector-ref vec 0))) 
         (kx (- (quotient krows 2) (- 1 (modulo krows 2))))
         (ky (- (quotient kcols 2) (- 1 (modulo kcols 2))))
         (inner (iota0 kcols))
         (iter (apply append (map (lambda (x) (map (lambda (y) (list x y))
                                                   inner))
                                  (iota0 krows)))))
    (define (image-ref-torus i j)
      (image-ref img (modulo i rows) (modulo j cols)))
    (define (reref i j x y)
      (list (+ i (- kx x)) (+ j (- ky y))))
    (define (vecref x y)
      (vector-ref (vector-ref vec x) y))
    (define (innersum i j x y)
      (* (apply image-ref-torus (reref i j x y))
         (vecref x y)))
    (make-image rows cols
               (lambda (i j)
                 (apply + (map (lambda (xy)
                                 (apply innersum i j xy))
                               iter))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (dilate img . array)
  (let* ((array (if (null? array) #(#(1 1)#(1 1)) (car array)))
         (tmp (convolve img array)))
    (make-image (image-rows tmp) (image-cols tmp)
                (lambda (i j)
                  (if (= 0 (image-ref tmp i j)) 0 1)))))

(define (erode img . array)
  (let* ((array (if (null? array) #(#(1 1)#(1 1)) (car array)))
         (sum (apply + (apply append (map vector->list (vector->list array)))))
         (tmp (convolve img array)))
    (make-image (image-rows tmp) (image-cols tmp)
                (lambda (i j)
                  (if (= sum (image-ref tmp i j)) 1 0)))))

(define (open img . array)
  (apply dilate (apply erode img array) array))

(define (close img . array)
  (apply erode (apply dilate img array) array))

(define (distance-transform img)
  (let ((maxval (* (image-rows img) (image-cols img)))
        (_dtm '((-1 -1 1.414) (-1 0 1) (-1 1 1.414)
                ( 0 -1 1    ) ( 0 0 0) ( 0 1 1    )
                ( 1 -1 1.414) ( 1 0 1) ( 1 1 1.414))))
    (define (loop img)
      (define (_dthelp i j)
        (define (_dtgetval x)
          (let* ((i_ (modulo (+ i (car x)) (image-rows img)))
                 (j_ (modulo (+ j (cadr x)) (image-cols img)))
                 (v (+ (caddr x) (image-ref img i_ j_))))
            (min v maxval)))
        (apply min (map _dtgetval _dtm)))
      (let ((new (make-image (image-rows img) (image-cols img) _dthelp)))
        (if (equal? img new)
            img
            (loop new))))
    (loop (image-map (lambda (x) (if (= 0 x) 0 maxval)) img))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Discrete Fourier Transform the slow old-fashioned way. :)
(define 2pi 360_deg)

(define (dft-helper f t)
  (let* ((bigN (vector-length f))
         (sqrtn (sqrt bigN))
         (in (iota0 bigN)))
    (define (inner m)
      (/ (apply + (map (lambda (n) (* (vector-ref f n)
                                      (exp (* t 0-i 2pi m n (/ bigN)))))
                       in))
         sqrtn))
    (list->vector (map inner in))))

(define (dft1d f) (dft-helper f 1))
(define (idft1d f) (dft-helper f -1))

(define (dft img)
  (image-map-cols dft1d (image-map-rows dft1d img)))

(define (idft img)
  (image-map-cols idft1d (image-map-rows idft1d img)))

; Fast Fourier Transform
; http://en.wikipedia.org/wiki/Cooley-Tukey_FFT_algorithm

(define (fft-help v t)
  (let ((n (vector-length v)))
    (if (= 1 n)
        v
        (let* ((vref (lambda (x) (vector-ref v x)))
               (n_2 (/ n 2))
               (in2 (iota0 n_2))
               (evn (list->vector
                     (map (lambda (x) (vref (* 2 x))) in2)))
               (odd (list->vector
                     (map (lambda (x) (vref (+ 1 (* 2 x)))) in2)))
               (fe (fft-help evn t))
               (fo (fft-help odd t))
               (feref (lambda (x) (vector-ref fe x)))
               (foref (lambda (x) (vector-ref fo x))))
          (list->vector
           (map (lambda (j)
                  (if (< j n_2)
                      (+ (feref j)
                         (* (exp (* t 0-i 2pi j (/ n)))
                            (foref j)))
                      (- (feref (- j n_2))
                         (* (exp (* t 0-i 2pi (- j n_2) (/ n)))
                            (foref (- j n_2))))))
                   (iota0 n)))))))

(define (fft-check v)
  (define (two-power? x)
    (or (= x 0) (= x 1) (let ((y (quotient x 2)))
                          (and (= x (* 2 y)) (two-power? y)))))
  (if (two-power? (vector-length v))
      v
      (error "FFT needs power-of-two size")))

(define (fft1d v)
  (let* ((v (fft-check v))
         (sq (sqrt (vector-length v))))
    (list->vector
     (map (lambda (x) (/ x sq)) (vector->list (fft-help v 1))))))

(define (ifft1d v)
  (let* ((v (fft-check v))
         (sq (sqrt (vector-length v))))
    (list->vector
     (map (lambda (x) (/ x sq)) (vector->list (fft-help v -1))))))

; FFT of 2d images
(define (image-map-rows f img)
  (vector->image
   (list->vector
    (map f (vector->list (image->vector img))))))

(define (image-map-cols f img)
  (image-transpose (image-map-rows f (image-transpose img))))

(define (fft img)
  (image-map-cols fft1d (image-map-rows fft1d img)))

(define (ifft img)
  (image-map-cols ifft1d (image-map-rows ifft1d img)))
